#include <catch.hpp>

#include "pls/internal/base/spin_lock.h"
#include "pls/internal/base/system_details.h"

#include <vector>
#include <mutex>
#include <thread>

using namespace pls::internal::base;

static bool base_tests_visited;
static int base_tests_local_value_one;
static std::vector<int> base_tests_local_value_two;

int base_tests_shared_counter;

TEST_CASE("spinlock protects concurrent counter", "[internal/data_structures/spinlock.h]") {
  constexpr int num_iterations = 1000000;
  base_tests_shared_counter = 0;
  spin_lock lock{};

  SECTION("lock can be used by itself") {
    std::thread t1{[&]() {
      for (int i = 0; i < num_iterations; i++) {
        lock.lock();
        base_tests_shared_counter++;
        lock.unlock();
      }
    }};
    std::thread t2{[&]() {
      for (int i = 0; i < num_iterations; i++) {
        lock.lock();
        base_tests_shared_counter--;
        lock.unlock();
      }
    }};

    t1.join();
    t2.join();

    REQUIRE(base_tests_shared_counter == 0);
  }

  SECTION("lock can be used with std::lock_guard") {
    std::thread t1{[&]() {
      for (int i = 0; i < num_iterations; i++) {
        std::lock_guard<spin_lock> my_lock{lock};
        base_tests_shared_counter++;
      }
    }};
    std::thread t2{[&]() {
      for (int i = 0; i < num_iterations; i++) {
        std::lock_guard<spin_lock> my_lock{lock};
        base_tests_shared_counter--;
      }
    }};

    t1.join();
    t2.join();

    REQUIRE(base_tests_shared_counter == 0);
  }
}
